@@ -21,24 +21,76 @@ with transaction(auto_apply=False) as ops:
 
 """
 
+import itertools
 from contextlib import contextmanager
+from contextvars import ContextVar
 from enum import Enum
 from typing import Callable, Generic, Optional, TypeVar
 
 try:
     from transaction import Transaction, TransactionManager  # noqa: WPS433 - nested import
     from transaction.interfaces import IDataManager  # noqa: WPS433 - nested import
+    from transaction._manager import _new_transaction  # noqa: WPS436, WPS450, WPS433 - nested import
     from zope.interface import implementer  # noqa: WPS433 - nested import
 except ImportError as exc:
     raise Exception('You need to install the `transaction` extra to use the transaction features') from exc
 
 
+txns = ContextVar('txns', default=[])
+id_counter = itertools.count()
+
+
+class TransactionExt(Transaction):
+    def _free(self):
+        # We don't want to free any resources after a nested transaction is commited or aborted.
+        # Otherwise the parent transaction won't be able to abort it.
+        pass
+
+
+class HierarchicalTransactionManager(TransactionManager):
+    @property
+    def transactions(self):
+        # We store transactions of a hierarchie in a ContextVar to make sure
+        # there are no conflicts in an async context.
+        return txns.get()
+
+    def __exit__(self, t, v, tb):
+        super().__exit__(t, v, tb)
+        self.transactions.pop()
+
+    def begin(self):
+        # We overload this method to remove a check and join child transactions
+        # to the parent transaction as a resource.
+        parent_transaction = self._txn
+        txn = self._txn = TransactionExt(self._synchs, self)  # noqa: WPS429
+        if parent_transaction is not None:
+            # We wrap the transaction to give it the same interface as a resource, and link it
+            # to the parent transaction. That way, the parent transaction will be able to abort its
+            # child transactions if any of them fails.
+            parent_transaction.join(TransactionResource(txn))
+        _new_transaction(txn, self._synchs)
+        return txn
+
+    @property
+    def _txn(self):
+        if self.transactions:
+            # Return the current transaction.
+            return self.transactions[-1]
+        return None
+
+    @_txn.setter
+    def _txn(self, value):
+        if value is not None:
+            self.transactions.append(value)
+
+
 @contextmanager
 def transaction(auto_apply: bool = True):
-    manager = TransactionManager(explicit=True)
+    manager = HierarchicalTransactionManager(explicit=True)
 
     with manager as txn:
-        ops = OperationManager(txn)
+        ops = OperationManager()
+        txn.join(ops)
 
         yield ops
 
@@ -101,9 +153,9 @@ class Operation(Generic[T]):
 
 @implementer(IDataManager)
 class OperationManager:  # noqa: WPS214 - too many methods
-    def __init__(self, txn: Transaction):
+    def __init__(self):
         self.operations = []
-        txn.join(self)
+        self._id = next(id_counter)
 
     def add(self, operation: Operation):
         self.operations.append(operation)
@@ -122,6 +174,9 @@ class OperationManager:  # noqa: WPS214 - too many methods
             if op.state != OperationState.applied:
                 raise OperationNotApplied(op.state)
 
+    def sortKey(self):  # noqa: N802
+        return self._id
+
     # Methods under this comment are required as the `IDataManager` interface doesn't define any behavior
     # https://zope.readthedocs.io/en/latest/zdgbook/ComponentsAndInterfaces.html#zope-components
 
@@ -136,3 +191,28 @@ class OperationManager:  # noqa: WPS214 - too many methods
 
     def commit(self, txn):
         pass
+
+
+@implementer(IDataManager)
+class TransactionResource:
+    def __init__(self, txn: Transaction):
+        self.txn = txn
+        self._id = next(id_counter)
+
+    def abort(self, txn: Transaction):
+        self.txn.abort()
+
+    def sortKey(self):  # noqa: N802
+        return self._id
+
+    def commit(self, txn: Transaction):
+        pass
+
+    def tpc_begin(self, txn):
+        pass
+
+    def tpc_vote(self, txn):
+        pass
+
+    def tpc_finish(self, txn):
+        pass
